#include <windows.h>
#include <process.h>
#include <assert.h>
#include <sys/timeb.h>

#include <string>

// ***********************************************************************
// Moment
// ***********************************************************************

class Moment {
public:
	void now () {
		_ftime(&tv);
	};

	double operator - ( Moment &m ) {
		double result = ((double)  tv.time - m.tv.time) * 1e3;
		result += ((double) tv.millitm - m.tv.millitm);
		return result;	};
private:
	struct _timeb tv;
};

// ***********************************************************************
// Thread
// ***********************************************************************

typedef unsigned int ( __stdcall *Thread_Main_T)(void *);

class Thread {
public:
	Thread ( Thread_Main_T p_main ) : thread_main(p_main) {};
	~Thread () {};

	uintptr_t run ( void *arg ) {
		unsigned thread_addr;
		hd = _beginthreadex(
			NULL, // Security attribute
			0, // Stack size as main thread
			thread_main, // Function to be called
			arg, // Optional argument
			0, // Thread should run immediately
			&thread_addr
			);
		return hd;
	};

	void join () {
		WaitForSingleObject(reinterpret_cast<HANDLE>(hd),INFINITE);
	};

	int getpid () const {
		return _getpid();
	};
private:
	unsigned thread_addr;
	uintptr_t hd;
	Thread_Main_T thread_main;
};

// ***********************************************************************
// Semaphore
// ***********************************************************************

class Semaphore {
public:
	Semaphore ( int p_initial, int p_max ) {
		s = CreateSemaphore(NULL,p_initial,p_max,NULL);
	};
	~Semaphore () {
		CloseHandle(s);
	};

	void P () { WaitForSingleObject(s,INFINITE); };
	void V () { ReleaseSemaphore(s,1,NULL); };

private:
	HANDLE s;
};

// ***********************************************************************
// SharedMemory
// ***********************************************************************

#include <iostream>

class SharedMemory {
public:
	SharedMemory ( const std::string &p_filename, int p_size )
		: filename(p_filename), len(p_size) {
		HANDLE pf = reinterpret_cast<HANDLE>(0xffffffff);
		shm = CreateFileMapping(
			pf, // Shared memory backed up by paging file
			NULL, // Security attributes
			PAGE_READWRITE,
			0,len, // 64 bit possible size of file mapping
			filename.c_str()
			);
		addr = MapViewOfFile(
			shm,
			FILE_MAP_ALL_ACCESS,
			0,0, // 64 bit offset
			len // size to be mapped, max. 32 bit
			);
	};
	
	SharedMemory ( const std::string &p_filename )
		: filename(p_filename) {
		shm = OpenFileMapping(
			FILE_MAP_ALL_ACCESS,
			FALSE, // inherit handle
			filename.c_str()
			);
		if (shm == NULL)
			std::cout << GetLastError() << std::endl;
		len = GetFileSize(shm,NULL);
		std::cout << len << std::endl;
		addr = MapViewOfFile(
			shm,
			FILE_MAP_ALL_ACCESS,
			0,0, // 64 bit offset
			len // size to be mapped, max. 32 bit
			);
	};

	~SharedMemory () {
		UnmapViewOfFile(addr);
		CloseHandle(shm);
	};

	char *base () const {
		return reinterpret_cast<char *>(addr);
	};

	unsigned long size () const {
		return len;
	};
private:
	std::string filename;
	HANDLE shm;
	DWORD len;
	void *addr;
};

// ***********************************************************************
// MappedFile
// ***********************************************************************

class MappedFile {
public:
	MappedFile ( std::string filename ) {
		fd = CreateFile(
			filename.c_str(),
			GENERIC_READ|GENERIC_WRITE,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL
			);
		len_low = GetFileSize(fd,&len_high);
		fdmap = CreateFileMapping(
			fd,
			NULL,
			PAGE_READWRITE,
			len_high,len_low,
			NULL
			);
		addr = MapViewOfFile(
			fdmap,
			FILE_MAP_ALL_ACCESS,
			0,0,
			len_low
			);
	};
	~MappedFile () {
		UnmapViewOfFile(addr);
		CloseHandle(fdmap);
		CloseHandle(fd);
	};

	char *base () const {
		return reinterpret_cast<char *>(addr);
	};

	int size () const {
		return len_low;
	};
private:
	HANDLE fd;
	HANDLE fdmap;
	void *addr;
	DWORD len_low;
	DWORD len_high;
};
